package katerlib;

import java.util.Vector;

import TUIO.TuioObject;
import TUIO.TuioProcessing;
import processing.core.PApplet;
import processing.core.PVector;

public class Kater implements IKaterEventDispatcher{
	private PApplet p;
	private TuioProcessing tuio;
	
	// config vars
    private int mySensorRadius;
    private int mySensorSpotsize;
    private int myAngleBlur;
    private int myDistanceBlur;
    private int myXoffset;
    private int myYoffset;
    private int myAngleOffset;
    private int myInnerRadius;
    private float movementAngle;
    
    private int myID;
    private float myX;
    private float myY;
    private float myAngle;
    private int myActionState = 0;
    private int myLastActionState = 0;
    private Boolean[] mySensorImage = {false, false, false, false};
    private PVector myVector = new PVector(0, 0);
    private PVector myTarget = new PVector(0, 0); 
    private Vector<IKaterEventListener> listenerList = new Vector<IKaterEventListener>();
    
    private Boolean debug = false;
    
    // activity vars
    private Boolean active = false;
    private Boolean running= false;
    
    /**
     * The single Kater object has to be created with a reference to the PApplet, the Tuio and the referencing tuioID
     * 
     * @param p The processing PApplet
     * @param tuio The TUIO Processing Client
     * @param theID	The TUIO ID of the Kater
     */
   public Kater(PApplet p, TuioProcessing tuio, int theID) {
		myID = theID;
		System.out.println("KaterLib: Kater No. "+myID+" created!");
		this.p=p;
		this.tuio=tuio;
	}
    
    void setValues(int theSensorRadius, int theSensorSpotsize, int theAngleBlur, int theDistanceBlur, int theXoffset, int theYoffset, int theAngleOffset, int theInnerRadius) {
        mySensorRadius = theSensorRadius;
        mySensorSpotsize = theSensorSpotsize;
        myAngleBlur = theAngleBlur;
        myDistanceBlur = theDistanceBlur;
        myXoffset = theXoffset;
        myYoffset = theYoffset;
        myAngleOffset = theAngleOffset;
        myInnerRadius = theInnerRadius;
    }
    /**
     * 
     * Send Coordinate to Kater by two floats
     * 
     * @param targetX
     * @param targetY
     */
    void goTo(float targetX, float targetY) {
    	this.running=true;
        myTarget.x = targetX;
        myTarget.y = targetY;
    }
    /**
     * 
     * Update the Kater each frame with his Coordiantes and Angle from TUIO
     * 
     * @param theX
     * @param theY
     * @param theAngle
     */
    public void update(float theX, float theY, float theAngle) {
        myX = theX;
        myY = theY;
        myVector.x = theX;
        myVector.y = theY;
        myAngle = theAngle+myAngleOffset;
        /*
         * Update the Status
         */
        movementAngle = myAngle + PApplet.degrees(PApplet.atan2(myVector.x - myTarget.x, myVector.y - myTarget.y));
        if (movementAngle > 180) {
            movementAngle -= 360;
        }
            
        if(running){
        if (myVector.dist(myTarget) > myDistanceBlur) {
        	
        	
        // check, if you are already there
            if ((movementAngle < myAngleBlur) && (movementAngle > -myAngleBlur)) {
            	
                // go ahead
                myActionState = 1;
            } else {
                if (movementAngle > myAngleBlur) {
                    // turn
                    myActionState = 3;
                } else if (movementAngle < -myAngleBlur) {
                    // turn
                    myActionState = 2;
                }
            }
        } else {
            myActionState = 0;
            dispatchFinish();
        }
 }else{	// not active
	    myActionState = 0;
        }
        if(myActionState!=myLastActionState){
        switch (myActionState) {
            case 0: // stop
                mySensorImage[0] = false;
                mySensorImage[1] = false;
                mySensorImage[3] = false;
                break;

            case 1: // forward
                mySensorImage[0] = true;
                mySensorImage[1] = false;
                mySensorImage[3] = false;
                break;

            case 2: // turn right
                mySensorImage[0] = false;
                mySensorImage[1] = true;
                mySensorImage[3] = false;
                break;

            case 3: // turn left
                mySensorImage[0] = false;
                mySensorImage[1] = false;
                mySensorImage[3] = true;
                break;
        }
        myLastActionState=myActionState;
        }
    }
    
    /**
     * 
     * Draw the Kater to the PApplet
     * 
     */
    void draw() {
    		p.pushMatrix();
    		p.translate(myX, myY);
        	p.ellipseMode(PApplet.CENTER);
            // Adjust the Sensor by 45 degrees

            p.rotate(PApplet.radians(135));
            p.noStroke();
            p.fill(0);
            // sensor fields to black
            p.ellipse(-mySensorRadius, 0, mySensorSpotsize, mySensorSpotsize);     // left
            p.ellipse(0, mySensorRadius, mySensorSpotsize, mySensorSpotsize);      // top
            p.ellipse(mySensorRadius, 0, mySensorSpotsize, mySensorSpotsize);      // right
            p.ellipse(0, -mySensorRadius, mySensorSpotsize, mySensorSpotsize);     // bottom
            // switching on sensor fields
            
            p.fill(255);
            
            if (mySensorImage[0]) {
                p.ellipse(-mySensorRadius, 0, mySensorSpotsize, mySensorSpotsize);     // left
            }
            
            if (mySensorImage[1]) {
                p.ellipse(0, -mySensorRadius, mySensorSpotsize, mySensorSpotsize);      // top
            }

            if (mySensorImage[2]) {
                p.ellipse(mySensorRadius, 0, mySensorSpotsize, mySensorSpotsize);      // right
            }

            if (mySensorImage[3]) {
                p.ellipse(0, mySensorRadius, mySensorSpotsize, mySensorSpotsize);     // bottom
            }
            
          
            p.fill(0);
            p.ellipse(0, 0, myInnerRadius, myInnerRadius);
            p.popMatrix();
            
            if(debug){
            	drawDebug();
            }
    	}
    
    /**
     * 
     *	draws the configuration and debug graphics 
     */
    private void drawDebug(){
    		 p.pushMatrix();
             p.fill(50);
             p.noStroke();
             p.ellipse(myTarget.x, myTarget.y, 10, 10);
             p.fill(255);
             p.stroke(50);
             p.strokeWeight(1);
             p.line(myTarget.x, myTarget.y, myVector.x, myVector.y);

             p.noStroke();
             p.text("Distance: " + myVector.dist(myTarget), mySensorRadius + 50 + myVector.x, myVector.y - 20);
             p.text("Angle: " + PApplet.degrees(PApplet.atan2(myVector.x - myTarget.x, myVector.y - myTarget.y)), mySensorRadius + 50 + myVector.x, myVector.y - 40);
             p.text("Object Angle: " + myAngle, mySensorRadius + 50 + myVector.x, myVector.y - 60);
             p.text("Movement Angle: " + movementAngle, mySensorRadius + 50 + myVector.x, myVector.y - 80);
             p.popMatrix();
    	
    }
    /**
     * 
     * @return the X Coordinate
     */
    public float getX(){return myX;}
    /**
     * 
     * @return the Y Coordinate
     */
    public float getY(){return myY;}
    /**
     * 
     * @return the Center X
     */
    public float getCenterX(){return (myX+myXoffset);}
    /**
     * 
     * @return the Center Y
     */
    public float getCenterY(){return (myY+myYoffset);}
    /**
     * 
     * @return the TUIO ID
     */
    public int getID(){return myID;}
    /**
     * set the running state to true
     * 
     */
    public void startrun(){running=true;}
    /**
     * 
     * set the running state to false and dispatch the finish event
     */
    void stop(){
    	running=false;
    	dispatchFinish();
    	}
    /**
     * 
     * set the Kater to an active state
     */
    void activate(){active=true;}
    /**
     * 
     * set the Kater to an deactive state
     */
    void deactivate(){active=false;}
    /**
     * 
     * @return if the Kater is currently running
     */
    Boolean getRunning(){return running;}
    /**
     * 
     * @return the TuioObject of the Kater
     */
    TuioObject getTobj(){
    	 Vector<TuioObject> tuioObjectList = tuio.getTuioObjects();
    	    for (int i = 0; i < tuioObjectList.size(); i++) {
    	        TuioObject tobj = (TuioObject) tuioObjectList.elementAt(i);
    	        if(tobj.getSymbolID()==myID){
    	        	return tobj;
    	        }
    	    }
    	    System.err.println("Kater TUIO Object not found");
    	    return null;
    	}
    /**
     * 
     * @return the angle
     */
    public float getAngle(){return myAngle;}

    /**
     * 
     * dispatch the finish event
     */
    public void dispatchFinish() {
    	this.running=false;
    	for (int i = 0; i < listenerList.size(); i++) {
    		listenerList.get(i).onActionStateChange(this, EKaterEventState.finished);
    	}
    }
    /**
     * 
     * turn the lights on the Kater on
     */
    public void lightsOn(){
    	mySensorImage[2] = true;
    }
    /**
     * 
     * turn the lights on the Kater off
     */
    public void lightsOff(){
    	mySensorImage[2] = false;
    }
    /**
     * add an event Listener
     * 
     * @param listener a Listener for Kater Eventdispatch
     */
	public void addActionListener(IKaterEventListener listener) {
		listenerList.add(listener);
		
	}
	/**
	 * remove an event Listener
	 * 
	 * @param listener a Listener for Kater Eventdispatch
	 */
	public void removeActionListener(IKaterEventListener listener) {
    	listenerList.remove(listener);
    }
	
	/**
	 * 
	 * @return the debug State
	 */
	public Boolean getDebug() {
		return debug;
	}
	/**
	 * sets the debug State
	 * @param debug Boolean if debug should be drawn or not
	 */
	public void setDebug(Boolean debug) {
		this.debug = debug;
	}
	/**
	 * 
	 * Toggles the Debug State
	 */
	public void toggleDebug(){
		debug=!debug;
	}
	
	
  }
